@@ -55,7 +55,7 @@
  */
 
 /*
- * $Id: ThreadTest.cpp,v 1.12 2001/06/22 14:22:00 tng Exp $
+ * $Id: ThreadTest.cpp,v 1.22 2002/08/26 14:47:49 tng Exp $
  *
  * @author Andy Heninger, IBM
  */
@@ -63,15 +63,15 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
-#include <parsers/SAXParser.hpp>
-#include <parsers/DOMParser.hpp>
-#include <util/PlatformUtils.hpp>
-#include <sax/HandlerBase.hpp>
-#include <framework/MemBufInputSource.hpp>
-
-#include <dom/DOM.hpp>
+#include <xercesc/parsers/SAXParser.hpp>
+#include <xercesc/parsers/XercesDOMParser.hpp>
+#include <xercesc/util/PlatformUtils.hpp>
+#include <xercesc/sax/HandlerBase.hpp>
+#include <xercesc/framework/MemBufInputSource.hpp>
 
+#include <xercesc/dom/DOM.hpp>
 
+void clearFileInfoMemory();
 
 //------------------------------------------------------------------------------
 //
@@ -85,7 +85,7 @@
 
 
 
-typedef void (*ThreadFunc)(void *);
+typedef DWORD (WINAPI *ThreadFunc)(void *);
 
 class ThreadFuncs           // This class isolates OS dependent threading
 {                           //   functions from the rest of ThreadTest program.
@@ -96,17 +96,30 @@ public:
 
 void ThreadFuncs::startThread(ThreadFunc func, void *param)
 {
-    unsigned long x;
-    x = _beginthread(func, 0x10000, param);
-    if (x == -1)
+    HANDLE  tHandle;
+    DWORD   threadID;
+
+    tHandle = CreateThread(0,          // Security Attributes,
+                           0x10000,    // Stack Size,
+                           func,       // Starting Address.
+                           param,      // Parmeters
+                           0,          // Creation Flags,
+                           &threadID); // Thread ID (Can not be null on 95/98)
+
+    if (tHandle == 0)
     {
         fprintf(stderr, "Error starting thread.  Errno = %d\n", errno);
+        clearFileInfoMemory();
         exit(-1);
     }
+
+    // Set the priority of the working threads low, so that the UI of the running system will
+    //   remain responsive.
+    SetThreadPriority(tHandle, THREAD_PRIORITY_IDLE);
 }
 
 
-#elif defined (AIX) || defined(SOLARIS) || defined(LINUX) || defined(HPUX) || defined (OS390)
+#elif defined (AIX) || defined(SOLARIS) || defined(LINUX) || defined(HPUX) || defined (OS390) || defined(FREEBSD) || defined(__CYGWIN__)
 #include <pthread.h>
 #include <unistd.h>
 #include <errno.h>
@@ -118,9 +131,7 @@ void ThreadFuncs::startThread(ThreadFunc func, void *param)
 //
 //------------------------------------------------------------------------------
 
-#ifdef OS390
 extern "C" {
-#endif
 
 
 typedef void (*ThreadFunc)(void *);
@@ -137,7 +148,12 @@ void ThreadFuncs::Sleep(int millis)
 {
    int seconds = millis/1000;
    if (seconds <= 0) seconds = 1;
+#if defined(SOLARIS)
+   // somehow the sleep hangs on Solaris
+   // so ignore the call
+#else
    ::sleep(seconds);
+#endif
 }
 
 
@@ -157,13 +173,12 @@ void ThreadFuncs::startThread(ThreadFunc func, void *param)
     if (x == -1)
     {
         fprintf(stderr, "Error starting thread.  Errno = %d\n", errno);
+        clearFileInfoMemory();
         exit(-1);
     }
 
 }
-#ifdef OS390
 }
-#endif
 #else
 #error This platform is not supported
 #endif
@@ -181,6 +196,8 @@ void ThreadFuncs::startThread(ThreadFunc func, void *param)
 struct InFileInfo
 {
     char    *fileName;
+    XMLCh   *uFileName;      // When doing an in-memory parse, avoid transcoding file name
+                             //    each time through.
     char    *fileContent;    // If doing an in-memory parse, this field points
                              //   to an allocated string containing the entire file
                              //   contents.  Otherwise it's 0.
@@ -204,12 +221,16 @@ struct RunInfo
 {
     bool        quiet;
     bool        verbose;
+    bool        stopNow;
     int         numThreads;
     bool        validating;
     bool        dom;
     bool        reuseParser;
     bool        inMemory;
     bool        dumpOnErr;
+    bool        doSchema;
+    bool        schemaFullChecking;
+    bool        doNamespaces;
     int         totalTime;
     int         numInputFiles;
     InFileInfo  files[MAXINFILES];
@@ -228,10 +249,12 @@ struct ThreadInfo
 {
     bool    fHeartBeat;            // Set true by the thread each time it finishes
                                    //   parsing a file.
+    bool    fInProgress;           // Set to false by the thread when parse in progress
     unsigned int     fParses;      // Number of parses completed.
     int              fThreadNum;   // Identifying number for this thread.
     ThreadInfo() {
         fHeartBeat = false;
+        fInProgress = false;
         fParses = 0;
         fThreadNum = -1;
     }
@@ -264,7 +287,8 @@ class ThreadParser: public HandlerBase
 private:
     int           fCheckSum;
     SAXParser*    fSAXParser;
-    DOMParser*    fDOMParser;
+    XercesDOMParser*    fXercesDOMParser;
+    DOMDocument * fDoc;
 
 
 public:                               //  This is the API used by the rest of the test program
@@ -274,14 +298,14 @@ public:                               //  This is the API used by the rest of th
     int parse(int fileNum);           // Parse the specified file.  fileNum is an index
                                       //   into the gRunInfo.files array.
                                       //  return the XML checksum, or
-                                      //  0 if a parse error occured.
+                                      //  0 if a parse error occurred.
 
     int reCheck();                    // Try to compute the checksum again.
                                       //  for DOM, re-walk the tree.
                                       //  for SAX, can't do, just return previous value.
 
-    void domPrint(const DOM_Node &node); // Dump out the contents of a node,
-    void domPrint();                   //   including any children.  Default (no param)
+    void domPrint(const DOMNode *node); // Dump out the contents of a node,
+    void domPrint();                  //   including any children.  Default (no param)
                                        //   version dumps the entire document.
 
 private:
@@ -289,7 +313,7 @@ private:
     const ThreadParser & operator =(const ThreadParser &); // No assignment.
 
     void  addToCheckSum(const XMLCh *chars, int len=-1);
-    void  domCheckSum(const DOM_Node &);
+    void  domCheckSum(const DOMNode *);
 
 
 public:                               // Not really public,
@@ -304,15 +328,15 @@ public:                               // Not really public,
 
     void warning(const SAXParseException& exception)     {
         fprintf(stderr, "*** Warning ");
-        throw;};
+        throw exception;};
 
     void error(const SAXParseException& exception)       {
         fprintf(stderr, "*** Error ");
-        throw;};
+        throw exception;};
 
     void fatalError(const SAXParseException& exception)  {
         fprintf(stderr, "***** Fatal error ");
-        throw;};
+        throw exception;};
 };
 
 
@@ -324,18 +348,25 @@ public:                               // Not really public,
 ThreadParser::ThreadParser()
 {
     fSAXParser = 0;
-    fDOMParser = 0;
+    fXercesDOMParser = 0;
+    fDoc       = 0;
     if (gRunInfo.dom) {
         // Set up to use a DOM parser
-        fDOMParser = new DOMParser;
-        fDOMParser->setDoValidation(gRunInfo.validating);
-        fDOMParser->setErrorHandler(this);
+        fXercesDOMParser = new XercesDOMParser;
+        fXercesDOMParser->setDoValidation(gRunInfo.validating);
+        fXercesDOMParser->setDoSchema(gRunInfo.doSchema);
+        fXercesDOMParser->setValidationSchemaFullChecking(gRunInfo.schemaFullChecking);
+        fXercesDOMParser->setDoNamespaces(gRunInfo.doNamespaces);
+        fXercesDOMParser->setErrorHandler(this);
     }
     else
     {
         // Set up to use a SAX parser.
         fSAXParser = new SAXParser;
         fSAXParser->setDoValidation(gRunInfo.validating);
+        fSAXParser->setDoSchema(gRunInfo.doSchema);
+        fSAXParser->setValidationSchemaFullChecking(gRunInfo.schemaFullChecking);
+        fSAXParser->setDoNamespaces(gRunInfo.doNamespaces);
         fSAXParser->setDocumentHandler(this);
         fSAXParser->setErrorHandler(this);
     }
@@ -347,7 +378,7 @@ ThreadParser::ThreadParser()
 ThreadParser::~ThreadParser()
 {
      delete fSAXParser;
-     delete fDOMParser;
+     delete fXercesDOMParser;
 }
 
 //------------------------------------------------------------------------
@@ -360,13 +391,14 @@ int ThreadParser::parse(int fileNum)
 {
     MemBufInputSource *mbis = 0;
     InFileInfo        *fInfo = &gRunInfo.files[fileNum];
+    bool              errors = false;
 
     fCheckSum = 0;
 
     if (gRunInfo.inMemory) {
         mbis = new  MemBufInputSource((const XMLByte *) fInfo->fileContent,
                                        fInfo->fileSize,
-                                       fInfo->fileName,
+                                       fInfo->uFileName,
                                        false);
     }
 
@@ -375,10 +407,10 @@ int ThreadParser::parse(int fileNum)
         if (gRunInfo.dom) {
             // Do a DOM parse
             if (gRunInfo.inMemory)
-                fDOMParser->parse(*mbis);
+                fXercesDOMParser->parse(*mbis);
             else
-                fDOMParser->parse(fInfo->fileName);
-            DOM_Document doc = fDOMParser->getDocument();
+                fXercesDOMParser->parse(fInfo->fileName);
+            DOMDocument* doc = fXercesDOMParser->getDocument();
             domCheckSum(doc);
         }
         else
@@ -396,10 +428,32 @@ int ThreadParser::parse(int fileNum)
         char *exceptionMessage = XMLString::transcode(e.getMessage());
         fprintf(stderr, " during parsing: %s \n Exception message is: %s \n",
             fInfo->fileName, exceptionMessage);
-        delete exceptionMessage;
+        delete [] exceptionMessage;
+        errors = true;
+    }
+    catch (const DOMException& toCatch)
+    {
+        fprintf(stderr, " during parsing: %s \n DOMException code is: %i \n",
+            fInfo->fileName, toCatch.code);
+        errors = true;
+    }
+    catch (const SAXParseException& e)
+    {
+        char *exceptionMessage = XMLString::transcode(e.getMessage());
+        fprintf(stderr, " during parsing: %s \n Exception message is: %s \n",
+            fInfo->fileName, exceptionMessage);
+        delete [] exceptionMessage;
+        errors = true;
+    }
+    catch (...)
+    {
+        fprintf(stderr, "Unexpected exception during parsing\n");
+        errors = true;
     }
 
     delete mbis;
+    if (errors)
+        return 0;  // if errors occurred, return zero as if checksum = 0;
     return fCheckSum;
 }
 
@@ -459,58 +513,58 @@ void ThreadParser::startElement(const XMLCh *const name, AttributeList &attribut
 // domCheckSum  -  Compute the check sum for a DOM node.
 //                 Works recursively - initially called with a document node.
 //
-void ThreadParser::domCheckSum(const DOM_Node &node)
+void ThreadParser::domCheckSum(const DOMNode *node)
 {
-    DOMString         s;
-    DOM_Node          child;
-    DOM_NamedNodeMap  attributes;
+    const XMLCh        *s;
+    DOMNode          *child;
+    DOMNamedNodeMap  *attributes;
 
-    switch (node.getNodeType() )
+    switch (node->getNodeType() )
     {
-    case DOM_Node::ELEMENT_NODE:
+    case DOMNode::ELEMENT_NODE:
         {
-            s = node.getNodeName();   // the element name
+            s = node->getNodeName();   // the element name
 
-            attributes = node.getAttributes();  // Element's attributes
-            int numAttributes = attributes.getLength();
+            attributes = node->getAttributes();  // Element's attributes
+            int numAttributes = attributes->getLength();
             int i;
             for (i=0; i<numAttributes; i++)
-                domCheckSum(attributes.item(i));
+                domCheckSum(attributes->item(i));
 
-            addToCheckSum(s.rawBuffer(), s.length());  // Content and Children
-            for (child=node.getFirstChild(); child!=0; child=child.getNextSibling())
+            addToCheckSum(s);          // Content and Children
+            for (child=node->getFirstChild(); child!=0; child=child->getNextSibling())
                 domCheckSum(child);
 
             break;
         }
 
 
-    case DOM_Node::ATTRIBUTE_NODE:
+    case DOMNode::ATTRIBUTE_NODE:
         {
-            s = node.getNodeName();  // The attribute name
-            addToCheckSum(s.rawBuffer(), s.length());
-            s = node.getNodeValue();  // The attribute value
+            s = node->getNodeName();  // The attribute name
+            addToCheckSum(s);
+            s = node->getNodeValue();  // The attribute value
             if (s != 0)
-                addToCheckSum(s.rawBuffer(), s.length());
+                addToCheckSum(s);
             break;
         }
 
 
-    case DOM_Node::TEXT_NODE:
-    case DOM_Node::CDATA_SECTION_NODE:
+    case DOMNode::TEXT_NODE:
+    case DOMNode::CDATA_SECTION_NODE:
         {
-            s = node.getNodeValue();
-            addToCheckSum(s.rawBuffer(), s.length());
+            s = node->getNodeValue();
+            addToCheckSum(s);
             break;
         }
 
-    case DOM_Node::ENTITY_REFERENCE_NODE:
-    case DOM_Node::DOCUMENT_NODE:
+    case DOMNode::ENTITY_REFERENCE_NODE:
+    case DOMNode::DOCUMENT_NODE:
         {
             // For entity references and the document, nothing is dirctly
             //  added to the checksum, but we do want to process the chidren nodes.
             //
-            for (child=node.getFirstChild(); child!=0; child=child.getNextSibling())
+            for (child=node->getFirstChild(); child!=0; child=child->getNextSibling())
                 domCheckSum(child);
             break;
         }
@@ -526,8 +580,7 @@ int ThreadParser::reCheck()
 {
     if (gRunInfo.dom) {
         fCheckSum = 0;
-        DOM_Document doc = fDOMParser->getDocument();
-        domCheckSum(doc);
+        domCheckSum(fDoc);
     }
     return fCheckSum;
 }
@@ -539,69 +592,76 @@ int ThreadParser::reCheck()
 //
 void ThreadParser::domPrint()
 {
-    DOMString("Begin DOMPrint ...\n").print();
+    printf("Begin DOMPrint ...\n");
     if (gRunInfo.dom)
-        domPrint(fDOMParser->getDocument());
-    DOMString("End DOMPrint\n").print();
+        domPrint(fXercesDOMParser->getDocument());
+    printf("End DOMPrint\n");
 }
 
-void ThreadParser::domPrint(const DOM_Node &node)
+static void printString(const XMLCh *str)
+{
+    char *s = XMLString::transcode(str);
+    printf("%s", s);
+    delete s;
+}
+
+
+void ThreadParser::domPrint(const DOMNode *node)
 {
 
-    DOMString         s;
-    DOM_Node          child;
-    DOM_NamedNodeMap  attributes;
+    DOMNode          *child;
+    DOMNamedNodeMap  *attributes;
 
-    switch (node.getNodeType() )
+    switch (node->getNodeType() )
     {
-    case DOM_Node::ELEMENT_NODE:
+    case DOMNode::ELEMENT_NODE:
         {
-            DOMString("<").print();
-            node.getNodeName().print();   // the element name
+            printf("<");
+            printString(node->getNodeName());   // the element name
 
-            attributes = node.getAttributes();  // Element's attributes
-            int numAttributes = attributes.getLength();
+            attributes = node->getAttributes();  // Element's attributes
+            int numAttributes = attributes->getLength();
             int i;
             for (i=0; i<numAttributes; i++) {
-                domPrint(attributes.item(i));
+                domPrint(attributes->item(i));
             }
-            DOMString(">").print();
+            printf(">");
 
-            for (child=node.getFirstChild(); child!=0; child=child.getNextSibling())
+            for (child=node->getFirstChild(); child!=0; child=child->getNextSibling())
                 domPrint(child);
 
-            DOMString("</").print();
-            node.getNodeName().print();
-            DOMString(">").print();
+            printf("</");
+            printString(node->getNodeName());
+            printf(">");
             break;
         }
 
 
-    case DOM_Node::ATTRIBUTE_NODE:
+    case DOMNode::ATTRIBUTE_NODE:
         {
-            DOMString(" ").print();
-            node.getNodeName().print();   // The attribute name
-            DOMString("= \"").print();
-            node.getNodeValue().print();  // The attribute value
-            DOMString("\"").print();
+            printf(" ");
+            printString(node->getNodeName());   // The attribute name
+            printf("= \"");
+            printString(node->getNodeValue());  // The attribute value
+            printf("\"");
             break;
         }
 
 
-    case DOM_Node::TEXT_NODE:
-    case DOM_Node::CDATA_SECTION_NODE:
+    case DOMNode::TEXT_NODE:
+    case DOMNode::CDATA_SECTION_NODE:
         {
-            node.getNodeValue().print();
+            printString(node->getNodeValue());
             break;
         }
 
-    case DOM_Node::ENTITY_REFERENCE_NODE:
-    case DOM_Node::DOCUMENT_NODE:
+    case DOMNode::ENTITY_REFERENCE_NODE:
+    case DOMNode::DOCUMENT_NODE:
         {
             // For entity references and the document, nothing is dirctly
             //  printed, but we do want to process the chidren nodes.
             //
-            for (child=node.getFirstChild(); child!=0; child=child.getNextSibling())
+            for (child=node->getFirstChild(); child!=0; child=child->getNextSibling())
                 domPrint(child);
             break;
         }
@@ -609,8 +669,6 @@ void ThreadParser::domPrint(const DOM_Node &node)
 }
 
 
-
-
 //----------------------------------------------------------------------
 //
 //   parseCommandLine   Read through the command line, and save all
@@ -629,6 +687,9 @@ void parseCommandLine(int argc, char **argv)
     gRunInfo.verbose = false;
     gRunInfo.numThreads = 2;
     gRunInfo.validating = false;
+    gRunInfo.doSchema = false;
+    gRunInfo.schemaFullChecking = false;
+    gRunInfo.doNamespaces = false;
     gRunInfo.dom = false;
     gRunInfo.reuseParser = false;
     gRunInfo.inMemory = false;
@@ -647,8 +708,27 @@ void parseCommandLine(int argc, char **argv)
                 gRunInfo.verbose = true;
             else if (strcmp(argv[argnum], "-v") == 0)
                 gRunInfo.validating = true;
-            else if (strcmp(argv[argnum], "-dom") == 0)
+            else if (strcmp(argv[argnum], "-s") == 0)
+                gRunInfo.doSchema = true;
+            else if (strcmp(argv[argnum], "-f") == 0)
+                gRunInfo.schemaFullChecking = true;
+            else if (strcmp(argv[argnum], "-n") == 0)
+                gRunInfo.doNamespaces = true;
+            else if (!strncmp(argv[argnum], "-parser=", 8)) {
+                const char* const parm = &argv[argnum][8];
+
+                if (!strcmp(parm, "dom")) {
+                    gRunInfo.dom = true;
+                }
+                else if (!strcmp(parm, "sax")) {
+                    gRunInfo.dom = false;
+                }
+                else
+                    throw 1;
+            }
+            else if (strcmp(argv[argnum], "-dom") == 0) {
                 gRunInfo.dom = true;
+            }
             else if (strcmp(argv[argnum], "-reuse") == 0)
                 gRunInfo.reuseParser = true;
             else if (strcmp(argv[argnum], "-dump") == 0)
@@ -705,15 +785,18 @@ void parseCommandLine(int argc, char **argv)
     catch (int)
     {
         fprintf(stderr, "usage:  threadtest [-v] [-threads nnn] [-time nnn] [-quiet] [-verbose] xmlfile...\n"
-            "     -v             Use validating parser.  Non-validating is default. \n"
-            "     -dom           Use a DOM parser.  Default is SAX. \n"
-            "     -quiet         Suppress periodic status display. \n"
-            "     -verbose       Display extra messages. \n"
+            "     -v             Use validating parser.  Non-validating is default.\n"
+            "     -n             Enable namespace processing. Defaults to off.\n"
+            "     -s             Enable schema processing. Defaults to off.\n"
+            "     -f             Enable full schema constraint checking. Defaults to off.\n"
+            "     -parser=xxx    Parser Type [dom | sax].  Default is SAX.\n"
+            "     -quiet         Suppress periodic status display.\n"
+            "     -verbose       Display extra messages.\n"
             "     -reuse         Retain and reuse parser.  Default creates new for each parse.\n"
-            "     -threads nnn   Number of threads.  Default is 2. \n"
+            "     -threads nnn   Number of threads.  Default is 2.\n"
             "     -time nnn      Total time to run, in seconds.  Default is forever.\n"
             "     -dump          Dump DOM tree on error.\n"
-            "     -mem           Read files into memory once only, and parse them from there.\n"
+            "     -mem           Read files into memory once only, and parse them from there.\n\n"
             );
         exit(1);
     }
@@ -741,9 +824,11 @@ void ReadFilesIntoMemory()
         for (fileNum = 0; fileNum <gRunInfo.numInputFiles; fileNum++)
         {
             InFileInfo *fInfo = &gRunInfo.files[fileNum];
+            fInfo->uFileName = XMLString::transcode(fInfo->fileName);
             fileF = fopen( fInfo->fileName, "rb" );
             if (fileF == 0) {
                 fprintf(stderr, "Can not open file \"%s\".\n", fInfo->fileName);
+                clearFileInfoMemory();
                 exit(-1);
             }
             fseek(fileF, 0, SEEK_END);
@@ -753,6 +838,7 @@ void ReadFilesIntoMemory()
             t = fread(fInfo->fileContent, 1, fInfo->fileSize, fileF);
             if (t != fInfo->fileSize) {
                 fprintf(stderr, "Error reading file \"%s\".\n", fInfo->fileName);
+                clearFileInfoMemory();
                 exit(-1);
             }
             fclose(fileF);
@@ -761,6 +847,21 @@ void ReadFilesIntoMemory()
     }
 }
 
+void clearFileInfoMemory()
+{
+    int     fileNum;
+
+    if (gRunInfo.inMemory)
+    {
+        for (fileNum = 0; fileNum <gRunInfo.numInputFiles; fileNum++)
+        {
+            InFileInfo *fInfo = &gRunInfo.files[fileNum];
+            delete [] fInfo->uFileName;
+            delete [] fInfo->fileContent;
+        }
+    }
+}
+
 
 
 //----------------------------------------------------------------------
@@ -769,19 +870,15 @@ void ReadFilesIntoMemory()
 //               Run in an infinite loop, parsing each of the documents
 //               given on the command line in turn.
 //
-//               There is no return from this fuction, and no graceful
-//               thread termination.  Threads are stuck running here
-//               until the OS shuts them down as a consequence of the
-//               main thread of the process (which never calls this
-//               function) exiting.
-//
 //----------------------------------------------------------------------
 
-#ifdef OS390
+#ifdef PLATFORM_WIN32
+unsigned long WINAPI threadMain (void *param)
+#else
 extern "C" {
-#endif
 
 void threadMain (void *param)
+#endif
 {
     ThreadInfo   *thInfo = (ThreadInfo *)param;
     ThreadParser *thParser = 0;
@@ -795,9 +892,11 @@ void threadMain (void *param)
     // Each time through this loop, one file will be parsed and its checksum
     // computed and compared with the precomputed value for that file.
     //
-    while (true)
+    while (gRunInfo.stopNow == false)
     {
 
+        thInfo->fInProgress = true;
+
         if (thParser == 0)
             thParser = new ThreadParser;
 
@@ -823,9 +922,12 @@ void threadMain (void *param)
             // Revisit - let the loop continue to run?
             int secondTryCheckSum = thParser->reCheck();
             fprintf(stderr, "   Retry checksum is %x\n", secondTryCheckSum);
-            if (gRunInfo.dumpOnErr)
-                thParser->domPrint();
+            if (gRunInfo.dumpOnErr) {
+               if (gRunInfo.dom)
+                  thParser->domPrint();
+            }
             fflush(stdout);
+            clearFileInfoMemory();
             exit(-1);
         }
 
@@ -838,12 +940,17 @@ void threadMain (void *param)
 
         thInfo->fHeartBeat = true;
         thInfo->fParses++;
+        thInfo->fInProgress = false;
     }
-}
 
-#ifdef OS390
+    delete thParser;
+#ifdef PLATFORM_WIN32
+    return 0;
+#else
+    return;
 }
 #endif
+}
 
 
 
@@ -902,7 +1009,7 @@ int main (int argc, char **argv)
 
         if (cksum == 0)
         {
-            fprintf(stderr, "An error occured while initially parsing %s\n",
+            fprintf(stderr, "An error occurred while initially parsing %s\n",
                 fileName);
             errors = true;
         };
@@ -910,19 +1017,28 @@ int main (int argc, char **argv)
         gRunInfo.files[n].checkSum = cksum;
         if (gRunInfo.verbose )
             printf("%x\n", cksum);
-        if (gRunInfo.dumpOnErr && errors)
-            mainParser->domPrint();
+        if (gRunInfo.dumpOnErr && errors) {
+           if (gRunInfo.dom)
+              mainParser->domPrint();
+        }
 
     }
-    if (errors)
+
+    delete mainParser;
+
+    if (errors) {
+        clearFileInfoMemory();
         exit(1);
+    }
 
     //
     //  Fire off the requested number of parallel threads
     //
 
-    if (gRunInfo.numThreads == 0)
+    if (gRunInfo.numThreads == 0) {
+        clearFileInfoMemory();
         exit(0);
+    }
 
     gThreadInfo = new ThreadInfo[gRunInfo.numThreads];
 
@@ -968,8 +1084,23 @@ int main (int argc, char **argv)
     //
     //  Time's up, we are done.  (We only get here if this was a timed run)
     //  Tally up the total number of parses completed by each of the threads.
-    //  To Do:  Run the main thread at higher priority, so that the worker threads
-    //    won't make much progress while we are adding up the results.
+    //
+    gRunInfo.stopNow = true;      // set flag, which will cause worker threads to stop.
+
+    //
+    //  Make sure all threads are done before terminate
+    //
+    for (threadNum=0; threadNum < gRunInfo.numThreads; threadNum++) {
+        while (gThreadInfo[threadNum].fInProgress == true) {
+            ThreadFuncs::Sleep(1000);
+        }
+        if (gRunInfo.verbose)
+            printf("Thread #%d: is finished.\n", threadNum);
+
+    }
+
+    //
+    //  We are done!   Count the number of parse and terminate the program
     //
     double totalParsesCompleted = 0;
     for (threadNum=0; threadNum < gRunInfo.numThreads; threadNum++)
@@ -978,12 +1109,18 @@ int main (int argc, char **argv)
         // printf("%f   ", totalParsesCompleted);
     }
 
-    double parsesPerMinute = totalParsesCompleted / (double(gRunInfo.totalTime) / double(60));
-    printf("\n%8.1f parses per minute.", parsesPerMinute);
+    if (gRunInfo.quiet == false) {
+        double parsesPerMinute = totalParsesCompleted / (double(gRunInfo.totalTime) / double(60));
+        printf("\n%8.1f parses per minute.\n", parsesPerMinute);
+    }
 
-    //  The threads are still running; we just return
-    //   and leave it to the operating sytem to kill them.
-    //
+    XMLPlatformUtils::Terminate();
+
+    clearFileInfoMemory();
+
+    delete [] gThreadInfo;
+
+    printf("Test Run Successfully\n");
 
     return 0;
 }
